/*
  Copyright 1997, 2002 Ronald S. Burkey.

  This file is part of Ron's Indexing Program (RIP).

  RIP is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  RIP is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GutenMark; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Filename:	rip_2.c 
  Target:	MS-DOS
  Compiler:	Borland Turbo C 2.0
  Purpose:	Code needed for RIP.C.
  Mods:		03/07/97 RSB	Split off from RIP.C because of compiler memory
		03/26/97 RSB	Allowed '\r' in place of '\n'.
		01/08/02 RSB	Reformatted in GNU style and otherwise
				cleaned up in preparation for first
				GPL'd release.
*/

#include "rip.h"

/* Read the code table.  The table is in the format output by HUFFMAN1.C. */
int
read_huffman (char *filename, int c_index)
{
  char s[257], ss[257], *sss;
  int i, j, k;
  unsigned long n;
  unsigned node, numd;
  COMPRESSOR *compress;
  DECOMPRESSOR *decompress;

  FILE *fin;
  fin = fopen (filename, "r");
  if (fin == NULL)
    return (1);
  for (; NULL != fgets (s, sizeof (s), fin);)
    {
      if (!strncmpi (s, "CODES:", 6))
	goto codes;
    }

huffman_error:

  fclose (fin);

  return (2);

  /* load the code table */
codes:
  compress = compressors[c_index];
  decompress = decompressors[c_index];
  (*decompress)[0].one = 0;
  (*decompress)[0].zero = 0;
  numd = 1;
  for (i = 0; i < 256; i++)
    {
      if (NULL == fgets (s, sizeof (s), fin))
	goto huffman_error;
      if (2 != sscanf (s, "%u:%s", &j, ss))
	goto huffman_error;
      if (j > 255)
	goto huffman_error;
      n = 0;
      k = 0;
      node = 0;
      for (sss = ss; *sss; sss++)
	{
	  n += n;
	  if (*sss == '1')
	    {
	      n++;
	      if ((*decompress)[node].one == 0)
		{
		  if (numd >= MAXD)
		    goto huffman_error;
		  (*decompress)[node].one = numd;
		  (*decompress)[numd].one = 0;
		  (*decompress)[numd].zero = 0;
		  numd++;
		}
	      node = (*decompress)[node].one;
	    }
	  else if (*sss == '0')
	    {
	      if ((*decompress)[node].zero == 0)
		{
		  if (numd >= MAXD)
		    goto huffman_error;
		  (*decompress)[node].zero = numd;
		  (*decompress)[numd].one = 0;
		  (*decompress)[numd].zero = 0;
		  numd++;
		}
	      node = (*decompress)[node].zero;
	    }
	  else
	    goto huffman_error;
	  k++;
	}
      (*decompress)[node].value = j;
      if (k > 32)
	goto huffman_error;
      (*compress)[j].numbits = k;
      (*compress)[j].bits = n;
    }
  fclose (fin);
  return (0);
}

/* Check if a word is a trash (unindexed) word or not */
int
iswimp (char *s)
{
  int j;
  if (strlen (s) >= MINSTRING)
    return (0);
  for (; *s; s++)
    {
      j = toupper (*s);
      if (j == 'Z' || j == 'J' || j == 'Q' || j == 'X')
	return (0);
    }
  return (1);
}

/* reads the header of a compressed file */
int
read_header (void)
{
  if (getc (fin) != 'R' ||
      getc (fin) != 'I' || getc (fin) != 'P' || getc (fin) != '3')
    return (1);
  if (1 != fread (&num_text, sizeof (num_text), 1, fin))
    return (2);
  if (1 != fread (&blocksize, sizeof (blocksize), 1, fin))
    return (3);
  blocks = (num_text + blocksize - 1) / blocksize;
  if (1 != fread (&index, sizeof (index), 1, fin))
    return (4);
  return (0);
}

/* create an output file */
int
create_file (char *fil)
{
  fout = fopen (fil, "r");
  if (fout != NULL)
    {
      fclose (fout);
      cprintf ("Output file %s already exists.\r\n", fil);
    }
  fout = fopen (fil, "w+b");
  if (fout == NULL)
    {
      fclose (fin);
      printf ("Cannot create output file %s.\n", fil);
      return (1);
    }
  setvbuf (fout, NULL, _IOFBF, 8192);
  return (0);
}

/* open the input file */
int
open_file (char *fil)
{
  fin = fopen (fil, "rb");
  if (fin == NULL)
    {
      printf ("Input file %s not found.\n", fil);
      return (1);
    }
  setvbuf (fin, NULL, _IOFBF, 16384);
  return (0);
}

/* Expand a rip-file back to original form. */
int
decompress_it (char *infile, char *outfile)
{
  int err;
  unsigned long n;
  unsigned i, k, node, kb;
  struct ftime ft;
  char *sss;
  DECOMPRESSOR *decompress;
  int lastchar, equal;

  equal = !strcmpi (infile, outfile);
  lastchar = ' ';
  kb = 0;
  for (sss = infile; *sss && *sss != '.'; sss++);
  if (strncmpi (sss, ".RIP", 4))
    return (3);

  for (sss = outfile; *sss && *sss != '.'; sss++);
  strcpy (sss, ".txt");
  if (equal)
    cprintf ("%s->TXT\r\n", infile);
  else
    cprintf ("%s->%s\r\n", infile, outfile);

  err = open_file (infile);
  if (err)
    return (0x10 + err);
  err = create_file (outfile);
  if (err)
    return (0x20 + err);
  err = read_header ();
  if (err)
    return (0x30 + err);
  n = num_text;

  if (fseek (fin, 4 * blocks, SEEK_CUR))
    return (1);

  i = 0;			/* packed input bits */
  k = 0;			/* bit count */
  for (; n; n--)
    {
      for (node = 0;;)
	{
	  if (!k)
	    {
	      i = getc (fin);
	      if (((int) i) == EOF)
		return (2);
	      k = 8;
	    }
	  if (isalpha (lastchar))
	    {
	      decompress = decompressors[toupper (lastchar) - 'A'];
	    }
	  else if (isdigit (lastchar))
	    {
	      decompress = decompressors[26];
	    }
	  else if (ispunctuation (lastchar))
	    {
	      decompress = decompressors[27];
	    }
	  else
	    {
	      decompress = decompressors[28];
	    }
	  if (i & 128)
	    node = (*decompress)[node].one;
	  else
	    node = (*decompress)[node].zero;
	  i = i << 1;
	  k--;
	  if (!(*decompress)[node].one && !(*decompress)[node].zero)
	    {
	      fputc (lastchar = (*decompress)[node].value, fout);
	      kb++;
	      if (kb == BLOCKSIZE)
		{
		  kb = 0;
		  lastchar = ' ';
		}
	      break;
	    }
	}
    }
  getftime (fileno (fin), &ft);
  fflush (fout);
  setftime (fileno (fout), &ft);
  if (ferror (fout))
    return (5);
  else
    {
      fclose (fout);
      fclose (fin);
      if (delete_file)
	remove (infile);
    }
  return (0);
}

/* Create a master index */
int
index_it (void)
{
  FILE *fin, /* *fout, */ *files;
  char s[257], ss[257], bits_per_char[256];
  unsigned long offset, n;
  unsigned u, i, fileno;
  HASHVALUE num_ndx, h, hk;
  NDX ndx;
  unsigned long t;

  t = clock ();
  for (u = 0; u < 256; u++)
    {
      i = 0;
      if (u & 1)
	i++;
      if (u & 2)
	i++;
      if (u & 4)
	i++;
      if (u & 8)
	i++;
      if (u & 16)
	i++;
      if (u & 32)
	i++;
      if (u & 64)
	i++;
      if (u & 128)
	i++;
      bits_per_char[u] = i;
    }
  files = fopen ("RIP.FIL", "w");
  fseek (ftemp, 0, SEEK_SET);
  fileno = -1;
  for (; NULL != gets (s);)
    {

      sscanf (s, "%s", ss);
      cprintf ("%s\r\n", ss);
      if (strlen (ss) > 60)
	{
	  cprintf ("Name too long.  Ignored.\n");
	  continue;
	}
      fin = fopen (ss, "rb");
      if (fin == NULL)
	{
	  printf ("Input file does not exist.\n");
	  continue;
	}
      fprintf (files, "%-60s\n", ss);
      fileno++;

      fseek (fin, 12, SEEK_SET);
      fread (&offset, sizeof (offset), 1, fin);
      fseek (fin, offset, SEEK_SET);
      fread (&num_ndx, sizeof (num_ndx), 1, fin);

      for (; num_ndx && 1 == fread (&ndx, sizeof (NDX), 1, fin); num_ndx--)
	{
	  if (ndx.hash)
	    {
	      fwrite (&fileno, sizeof (fileno), 1, ftemp);
	      fwrite (&ndx.hash, sizeof (ndx.hash), 1, ftemp);
	    }
	}

      fclose (fin);
    }
  fclose (files);
  n = ftell (ftemp) / (sizeof (unsigned) + sizeof (HASHVALUE));
  fseek (ftemp, 0, SEEK_SET);
  cprintf ("Sorting master index ... ");
  {
    int cmp_hash (void *, void *);
    fclose (ftemp);
    i = longsort4 (cmp_hash,
		   "RIP.TMP",
		   sizeof (unsigned) + sizeof (HASHVALUE), 0, n, "RIP3.TMP");
    if (i)
      {
	cprintf ("\r\nError code=%u\r\n", i);
	exit (1);
      }
    ftemp = fopen ("RIP3.TMP", "r+b");
    setvbuf (ftemp, NULL, _IOFBF, 16384);
  }
  cprintf ("and saving it.\r\n");
  fseek (ftemp, 0, SEEK_END);
  n =
    (ftell (ftemp) / (sizeof (unsigned) + sizeof (HASHVALUE))) *
    sizeof (unsigned) + 4;
  fout = fopen ("RIP.IND", "wb");
  setvbuf (fout, NULL, _IOFBF, 16384);
  fwrite (&n, sizeof (n), 1, fout);
  fseek (ftemp, 0, SEEK_SET);
  n = (n - 4) / sizeof (unsigned);
  hk = 0;
  offset = ftell (fout);
  fseek (ftemp2, 0, SEEK_SET);
  while (n)
    {
      fread (&i, sizeof (i), 1, ftemp);
      fread (&h, sizeof (h), 1, ftemp);
      if (hk != h && hk)
	{
	  fwrite (&hk, sizeof (hk), 1, ftemp2);
	  fwrite (&offset, sizeof (offset), 1, ftemp2);
	  offset = ftell (fout);
	}
      hk = h;
      fwrite (&i, sizeof (i), 1, fout);
      n--;
    }
  fwrite (&hk, sizeof (hk), 1, ftemp2);
  fwrite (&offset, sizeof (offset), 1, ftemp2);
  hk = 0;
  offset = ftell (fout);
  fwrite (&hk, sizeof (hk), 1, ftemp2);
  fwrite (&offset, sizeof (offset), 1, ftemp2);
#pragma warn -sig
  h = ftell (ftemp2) / (sizeof (long) + sizeof (HASHVALUE));
#pragma warn +sig
  fwrite (&h, sizeof (h), 1, fout);
  fseek (ftemp2, 0L, SEEK_SET);
  while (h)
    {
      fread (&hk, sizeof (hk), 1, ftemp2);
      fread (&n, sizeof (n), 1, ftemp2);
      fwrite (&hk, sizeof (hk), 1, fout);
      fwrite (&n, sizeof (n), 1, fout);
      h--;
    }
  fclose (fout);
#ifndef LONGSORT3
  fclose (ftemp);
  remove ("RIP3.TMP");
  ftemp = fopen ("RIP.TMP", "r+b");
  setvbuf (ftemp, NULL, _IOFBF, 16384);
#endif
  cprintf ("Done, %lu ticks.\r\n", clock () - t);
  return (0);
}

int
ispunctuation (char c)
{
  if (c == '.')
    return (1);
  if (c == '!')
    return (1);
  if (c == '?')
    return (1);
  if (c == ':')
    return (1);
  if (c == ';')
    return (1);
  if (c == ',')
    return (1);
  if (c == ')')
    return (1);
  return (0);
}

/* find hash entry in a sorted table */
unsigned
lookup (unsigned v)
{
  unsigned mn, mx, m, u;
  unsigned long mm;
  mn = 0;
  mx = table_entries;
  for (;;)
    {
      if (mx <= mn)
	return (0xffffU);
      mm = mn;
#pragma warn -sig
      m = (mm + mx) / 2;
#pragma warn +sig
      u = file_hashes.ib[m];
      if (u == v)
	return (m);
      else if (u < v)
	mn = m + 1;
      else
	mx = m;
    }
}

/* Same principle, but we have to load the table, of type NDX[], from
   disk.  */
unsigned
lookup_ndx (HASHVALUE v, NDX * ndx, unsigned mx)
{
  HASHVALUE u;
  unsigned m, mn;
  unsigned long mm;
  mn = 0;
  if (v < ndx[0].hash)
    return (0xffffU);
  if (v > ndx[mx - 1].hash)
    return (0xffffU);
  for (;;)
    {
      if (mx <= mn)
	return (0xffffU);
      mm = mn;
#pragma warn -sig
      m = (mm + mx) / 2;
#pragma warn +sig
      u = ndx[m].hash;
      if (u == v)
	return (m);
      else if (u < v)
	mn = m + 1;
      else
	mx = m;
    }
}


/*
  Note that the hash index may be too large to load all at one time, so we
  may need to use smaller chunks of it.  It is assumed that we have already
  seeked to the proper location in the file.

  Note also that the terminator record written at the end of the hash
  index is included in the count, but that it is out of order with the
  others since its hash value is zero.
*/
#define NEW_INDEX_SEARCH
unsigned next_search, max_search;
unsigned
findhash (FILE * f)
{
  HASHVALUE num_hashes, mx;
  unsigned block;
  unsigned i, j, byte, mask, found;
  unsigned long n;
  unsigned long index_start;
  HASHVALUE mn, m;
  NDX current_ndx;

  block = 0xffffU;
  max_search = next_search = 0;
  for (i = 0; i < 0x2000; i++)
    file_block[i] = 0xff;

  fread (&num_hashes, sizeof (num_hashes), 1, f);
  if (num_hashes == 0 || num_hashes == -1)
    goto done;

  index_start = ftell (f);
  for (found = 2; found < num_search; found++)
    if (search[found].hash)
      {
	mn = 0;
	mx = num_hashes;
	for (; mx > mn;)
	  {
	    m = (mx + mn) / 2;
	    fseek (f, index_start + sizeof (NDX) * m, SEEK_SET);
	    fread (&current_ndx, sizeof (NDX), 1, f);
	    if (search[found].hash < current_ndx.hash)
	      mx = m;
	    else if (search[found].hash > current_ndx.hash)
	      mn = m + 1;
	    else
	      {
		search[found].start_offset = current_ndx.offset;
		fread (&current_ndx, sizeof (NDX), 1, f);
		search[found].end_offset = current_ndx.offset;
		break;
	      }
	  }
	if (mx <= mn)
	  break;
      }
    else
      found++;

  /* What we do now is to set up the file_block array so that every
     block found is in it. */
  if (found >= num_search)
    {
      for (i = 2; i < num_search; i++)
	{
	  block = 0;
	  if (search[i].hash)
	    {
	      n = search[i].start_offset;
	      fseek (f, n, SEEK_SET);
	      while (n < search[i].end_offset)
		{
		  fread (&j, sizeof (j), 1, f);
		  if (j > max_search)
		    max_search = j;
		  for (; block < j; block++)
		    {
		      byte = block >> 3;
		      mask = 1 << (block & 7);
		      file_block[byte] &= ~mask;
		    }
		  block = j + 1;
		  n += 2;
		}
	      for (; block < 0xffffU; block++)
		{
		  byte = block >> 3;
		  mask = 1 << (block & 7);
		  file_block[byte] &= ~mask;
		}
	    }
	}
      block = nexthash ();
    }

done:
  return (block);
}

unsigned
nexthash (void)
{
  unsigned byte, mask;
  for (; next_search <= max_search; next_search++)
    {
      byte = next_search >> 3;
      mask = 1 << (next_search & 7);
      if (file_block[byte] & mask)
	break;
    }
  if (next_search <= max_search)
    return (next_search++);
  return (0xffffU);
}

/* Uncompress next byte in a file. The g_xxx variables are assumed to have
   been set up properly in advance, possibly by get_text().  At the
   beginning of a file, that appropriate values are the ones shown in
   the initializers.  Returns -1 on error. */
char g_lastchar = ' ';
unsigned g_kb = 0, g_bits = 0, g_data;
int lastchar, ignore_cr;
int
get_onebyte (FILE * fin)
{
  unsigned j, node;
  DECOMPRESSOR *decompress;

  for (node = 0;;)
    {
      if (!g_bits)
	{
	  g_data = getc (fin);
	  if (((int) g_data) == EOF)
	    return (-1);
	  g_bits = 8;
	}
      if (isalpha (g_lastchar))
	{
	  decompress = decompressors[toupper (g_lastchar) - 'A'];
	}
      else if (isdigit (g_lastchar))
	{
	  decompress = decompressors[26];
	}
      else if (ispunctuation (g_lastchar))
	{
	  decompress = decompressors[27];
	}
      else
	{
	  decompress = decompressors[28];
	}
      if (g_data & 128)
	node = (*decompress)[node].one;
      else
	node = (*decompress)[node].zero;
      g_data = g_data << 1;
      g_bits--;
      if (!(*decompress)[node].one && !(*decompress)[node].zero)
	{
	  j = g_lastchar = (*decompress)[node].value;
	  g_kb++;
	  if (g_kb == BLOCKSIZE)
	    {
	      g_kb = 0;
	      g_lastchar = ' ';
	    }
	  ignore_cr = (lastchar == '\r');
	  lastchar = j;
	  return (j);
	}
    }
}

/* Uncompress text beginning at a given address in the uncompressed file.
   "num" is the desired number of bytes, while the actual number uncompressed
   is returned.  In general, the number uncompressed is always equal to the
   number requested unless the end of file is reached.  */
unsigned
get_text (FILE * fin, unsigned long num_text, unsigned long loffset,
	  char *text_buf, unsigned num)
{
  unsigned long n;
  unsigned block, offset, j;
  int i;

  lastchar = 0;
  ignore_cr = 0;

#pragma warn -sig
  block = loffset / BLOCKSIZE;
  offset = loffset % BLOCKSIZE;
#pragma warn +sig

  fseek (fin, 16 + 4 * block, SEEK_SET);
  fread (&n, sizeof (n), 1, fin);
  g_lastchar = ' ';
  g_kb = 0;
#pragma warn -sig
  g_bits = (n & 7);
#pragma warn +sig
  n /= 8;
  fseek (fin, n, SEEK_SET);
  g_data = fgetc (fin) << g_bits;
  g_bits = 8 - g_bits;
#pragma warn -sig
  if (loffset >= num_text)
    {
      num = 0;
      goto done;
    }
  if (loffset + num > num_text)
    num = num_text - loffset;
#pragma warn +sig
  for (; offset; offset--)
    if (-1 == get_onebyte (fin))
      {
	num = 0;
	goto done;
      }
  for (j = 0; j < num; j++)
    {
      i = get_onebyte (fin);
      if (i == -1)
	{
	  num = j;
	  goto done;
	}
      text_buf[j] = i;
    }
done:
  return (num);
}

/* match a word against the search group */
int matchboth = 1;
int
wordmatch (char *s)
{
  int i;
  if (matchboth)
    {
      for (i = 2; i < num_search; i++)
	if (!strcmpi (s, search[i].string))
	  return (i);
    }
  else
    {
      for (i = 0; i < num_searchwords; i++)
	if (!strcmpi (s, searchwords[i]))
	  return (1);
    }
  return (0);
}

int
disp_line (void)
{
  char word[81], shortword[81];
  int word_ptr, short_ptr;
  char *ss;
  lowvideo ();
  short_ptr = word_ptr = 0;
  /* cprintf("%s",s); */
  for (ss = s; *ss && wherex () < 80; ss++)
    {
      if (isspace (*ss))
	{
	disp_word:
	  if (word_ptr)
	    {
	      word[word_ptr] = 0;
	      shortword[short_ptr] = 0;
	      word_ptr = 0;
	      short_ptr = 0;
	      if (wordmatch (shortword))
		highvideo ();
	      cprintf ("%s", word);
	    }
	  lowvideo ();
	  if (*ss)
	    putch (*ss);
	  else
	    break;
	}
      else
	{
	  if (word_ptr < 79)
	    word[word_ptr++] = *ss;
	  if (short_ptr < 79 && isalpha (*ss))
	    shortword[short_ptr++] = *ss;
	}
    }
  if (word_ptr)
    {
      *ss = 0;
      goto disp_word;
    }
  if (wherey () < 25)
    {
      cprintf ("\r\n");
      clreol ();
      return (0);
    }
  else
    return (1);
}

/* gets next line from ripfile.  returns -1 if EOF reached. */
int overflow_oneline, reenter_oneline = 0, reentry_char;
unsigned long current, firstline, nextline;
char *ss_oneline;
int
get_oneline (FILE * fin)
{
  int c;
  int retval;

  retval = 0;
  c = reentry_char;
  if (reenter_oneline == 1)
    goto first_reentry;
  else if (reenter_oneline == 2)
    goto second_reentry;
  else if (reenter_oneline == 3)
    goto third_reentry;
  else if (reenter_oneline == 4)
    goto fourth_reentry;
  else if (reenter_oneline == 5)
    goto fifth_reentry;

  for (;;)
    {

      c = get_onebyte (fin);
      current++;
      if (current > num_text)
	{
	  c = '\n';
	  retval = -1;
	  current = num_text;
	  if (firstline == 0xffffffffL)
	    firstline = current;
	}

      if (c == -1)
	{
	  reenter_oneline = 3;
	  nextline = current;
	  return (-1);
	third_reentry:
	  ss_oneline = s;
	  *s = 0;
	  return (-1);
	}
      else if (c == '\n' && ignore_cr)
	{
	}
      else if (c == '\n' || c == '\f' || c == '\r')
	{
	  reenter_oneline = 5;
	  nextline = current;
	  return (retval);
	fifth_reentry:
	  ss_oneline = s;
	  *s = 0;
	  if (firstline == 0xffffffffL)
	    firstline = current;
	}
      else if (c == '\b')
	{
	  if (ss_oneline != s)
	    {
	      ss_oneline--;
	      *ss_oneline = 0;
	    }
	}
      else if (c == '\t')
	{
	  if (!overflow_oneline)
	    {
#pragma warn -sig
	      c = ss_oneline - s;
#pragma warn +sig
	      for (c = ((c + 8) & 0xf8) - c; c; c--)
		*ss_oneline++ = ' ';
	      *ss_oneline = 0;
	    }
	}
      else
	{
	  if (isspace (c))
	    {
	      if (!overflow_oneline)
		{
		  *ss_oneline++ = c;
		  *ss_oneline = 0;
		}
	    }
	  else
	    {
	      if (!isprint (c))
		c = '.';
	      if (overflow_oneline)
		{
		  overflow_oneline = 0;
		  if (firstline == 0xffffffffL)
		    firstline = current - 1;
		}
	      *ss_oneline++ = c;
	      *ss_oneline = 0;
	    }
	}
      if (ss_oneline - s > 79)
	{
	  if (isspace (s[79]))
	    {
	      overflow_oneline = 1;
	      s[79] = 0;
	      reenter_oneline = 4;
	      nextline = current;
	      return (0);
	    fourth_reentry:
	      ss_oneline = s;
	      *s = 0;
	    }
	  else
	    {
	      for (c = 78; c >= 0; c--)
		if (isspace (s[c]))
		  break;
	      if (c < 0)
		{
		  c = s[79];
		  s[79] = 0;
		  if (firstline == 0xffffffffL)
		    firstline = current - 1;
		  reentry_char = c;
		  reenter_oneline = 1;
		  nextline = current;
		  return (0);
		first_reentry:
		  s[0] = c;
		  ss_oneline = s + 1;
		  *ss_oneline = 0;
		}
	      else
		{
		  s[c] = 0;
		  nextline = current - (79 - c);
		  if (firstline == 0xffffffffL)
		    firstline = nextline;
		  reentry_char = c;
		  reenter_oneline = 2;
		  return (0);
		second_reentry:
		  strcpy (s, s + c + 1);
		  ss_oneline = s + 79 - c;
		  *ss_oneline = 0;
		}
	    }
	}
    }
}

void
init_oneline (void)
{
  *s = 0;
  ss_oneline = s;
  overflow_oneline = 0;
  reenter_oneline = 0;
}

/* browse a compressed file, beginning at a given offset */
char searchstring[101], searchdummy[101];
int
browse (char *infile, FILE * fin,
	unsigned long num_text, unsigned long offset)
{
  int c, i, j, exactmatch;
  unsigned long newline, prior;
#define MAX_BACKSTEPS 1024
  unsigned long backsteps[MAX_BACKSTEPS];
  unsigned num_backsteps;
  unsigned backstep;
  unsigned long nn, nnn;
#define MAX_WORDLEN 79
  char word[MAX_WORDLEN + 1];
  int on_word;
  char *ss;
  FILE *bookmark;

  exactmatch = 0;
  matchboth = 1;
  prior = 0xffffffffL;
  /* view the find */
  lowvideo ();
#define PERPAGE 23
  textcolor (CLR_COMMAND);
  textbackground (BCK_COMMAND);
  clrscr ();

  highvideo ();
  cprintf ("COMMANDS:  Q");
  lowvideo ();
  cprintf ("uit, ");

  highvideo ();
  putch ('.');
  lowvideo ();
  putch ('/');
  highvideo ();
  putch ('+');
  lowvideo ();
  putch ('/');
  highvideo ();
  putch ('-');
  lowvideo ();
  putch ('/');
  highvideo ();
  cprintf ("Alt-S");
  lowvideo ();
  cprintf (" 1st/next/prior/save file, ");

  highvideo ();
  cprintf ("PgUp");
  lowvideo ();
  putch ('/');
  highvideo ();
  cprintf ("PgDn");
  lowvideo ();
  putch ('/');
  highvideo ();
  cprintf ("Up");
  lowvideo ();
  putch ('/');
  highvideo ();
  cprintf ("Down");
  lowvideo ();
  putch ('/');
  highvideo ();
  cprintf ("Home");
  lowvideo ();
  highvideo ();
  cprintf ("\r\n");

  cprintf ("FILE %s\r\n", infile);
jump:
  prior = offset;
  get_text (fin, num_text, offset, NULL, 0);
  for (;;)
    {
      textcolor (CLR_TEXT);
      textbackground (BCK_TEXT);
      lowvideo ();
      gotoxy (1, 3);
      clreol ();
      firstline = 0xffffffffL;

      current = offset;
      init_oneline ();
      for (i = 2; i < 25; i++)
	{
	  c = get_oneline (fin);
	  disp_line ();
	  if (c == -1)
	    {
	      while (wherey () < 25)
		{
		  cprintf ("\r\n");
		  clreol ();
		}
	      i = 25;
	    }
	}

      textcolor (CLR_COMMAND);
      textbackground (BCK_COMMAND);
      highvideo ();
      gotoxy (56, 2);
      cprintf ("OFFSET: %lu", offset);
      clreol ();

      lowvideo ();
      for (;;)
	{
	  struct time t, last_t;
	  gettime (&last_t);
	show_time:
	  gotoxy (46, 2);
	  textcolor (CLR_COMMAND);
	  textbackground (BCK_COMMAND);
	  lowvideo ();
	  cprintf ("%02d:%02d:%02d ",
		   last_t.ti_hour, last_t.ti_min, last_t.ti_sec);
	  gotoxy (80, 2);
	  while (!kbhit ())
	    {
	      gettime (&t);
	      if (t.ti_sec != last_t.ti_sec)
		{
		  last_t = t;
		  goto show_time;
		}
	    }
	  c = getch ();
	  if (c == ' ')
	    goto pgdn;
	  if (c == 0)
	    {
	      c = getch ();
	      if (c == N_PGUP)
		{
		  backstep = 24;
		backstep:
		  if (offset == 0)
		    continue;
		  if (offset < BLOCKSIZE)
		    nn = 0;
		  else
		    nn = offset - BLOCKSIZE;
		  nnn = nn;
		  get_text (fin, num_text, nn, NULL, 0);
		  init_oneline ();
		  current = nn;
		  nn = offset;
		  for (nextline = 0, num_backsteps = 0; nextline < nn;)
		    {
		      get_oneline (fin);
		      backsteps[num_backsteps++] = nextline;
		    }
		  if (num_backsteps >= backstep)
		    offset = backsteps[num_backsteps - backstep];
		  else if (nnn && num_backsteps)
		    offset = backsteps[0];
		  else
		    offset = 0;
		  goto jump;
		}
	      else if (c == N_PGDN)
		{
		pgdn:
		  offset = nextline;
		  goto jump;
		}
	      else if (c == N_HOME)
		{
		  offset = 0;
		  goto jump;
		}
	      else if (c == N_END)
		{
		  offset = num_text;
		  goto jump;
		}
	      else if (c == N_DOWN)
		{
		  offset = firstline;
		  goto jump;
		}
	      else if (c == N_UP)
		{
		  backstep = 2;
		  goto backstep;
		}
	      else if (c == A_S)
		{
		  c |= 0x80;
		  goto done;
		}
	    }
	  else
	    {
	      c = toupper (c);
	      if (c == 'S')
		{
		  exactmatch = 1;
		  gotoxy (1, 3);
		  textcolor (CLR_COMMAND);
		  textbackground (BCK_COMMAND);
		  highvideo ();
		  cprintf ("Search string: ");
		  clreol ();
		  gets (searchstring);
		  strupr (searchstring);
		  textcolor (CLR_TEXT);
		  textbackground (BCK_TEXT);
		  lowvideo ();
		  goto find_it;
		}
	      else if (c == 'F')
		{
		  exactmatch = 0;
		  matchboth = 0;
		  gotoxy (1, 3);
		  textcolor (CLR_COMMAND);
		  textbackground (BCK_COMMAND);
		  highvideo ();
		  cprintf ("Find string: ");
		  clreol ();
		  gets (s);
		  for (ss = s; *ss;)
		    if (isspace (*ss) || isalpha (*ss))
		      ss++;
		    else
		      strcpy (ss, ss + 1);
		  num_searchwords =
		    sscanf (s, "%s %s %s %s %s %s %s %s %s %s",
			    searchwords[0], searchwords[1], searchwords[2],
			    searchwords[3], searchwords[4], searchwords[5],
			    searchwords[6], searchwords[7], searchwords[8],
			    searchwords[9]);
		  textcolor (CLR_TEXT);
		  textbackground (BCK_TEXT);
		  lowvideo ();
		  goto find_it;
		}
	      else if (c == 'N')
		{
		find_it:
		  if (exactmatch)
		    {
		      gotoxy (56, 2);
		      i = strlen (searchstring);
		      cprintf ("SEARCHING ...");
		      clreol ();
		      get_text (fin, num_text, firstline, NULL, 0);
		      newline = current = firstline;
		      on_word = 0;
		      for (;;)
			{
			  if (on_word >= i)
			    {
			      if (!strncmp (searchdummy, searchstring, i))
				{
				  offset = newline;
				  goto jump;
				}
			      else
				{
				  for (j = 1; j < i; j++)
				    if (!strncmp
					(searchdummy + j, searchstring,
					 i - j))
				      break;
				  if (j < i)
				    strncpy (searchdummy, searchdummy + j,
					     on_word - j);
				  on_word -= j;
				}
			    }
			  c = get_onebyte (fin);
			  if (c == -1)
			    {
			      putchar ('\a');
			      goto jump;
			    }
			  current++;
			  if ((c == '\n' && !ignore_cr) || c == '\r')
			    newline = current;
			  searchdummy[on_word++] = toupper (c);
			}
		    }
		  else
		    {
		      gotoxy (56, 2);
		      cprintf ("FINDING ...");
		      clreol ();
		      get_text (fin, num_text, firstline, NULL, 0);
		      newline = current = firstline;
		      for (;;)
			{
			  c = get_onebyte (fin);
			  current++;
			check_next_word:
			  if (c == -1)
			    {
			      putchar ('\a');
			      goto jump;
			    }
			  if ((c == '\n' && !ignore_cr) || c == '\r')
			    newline = current;
			  if (!isalpha (c))
			    continue;
			  word[0] = c;
			  on_word = 1;
			  for (;;)
			    {
			      c = get_onebyte (fin);
			      current++;
			      if ((c == '\n' && !ignore_cr) || c == '\r')
				newline = current;
			      if (isalpha (c))
				{
				  if (on_word < MAX_WORDLEN)
				    word[on_word++] = c;
				}
			      else
				{
				  word[on_word++] = 0;
				  if (wordmatch (word))
				    {
				      offset = newline;
				      goto jump;
				    }
				  goto check_next_word;
				}
			    }
			}
		    }
		}
	      else if (c == 'P')
		{
		  if (prior != 0xffffffffL)
		    {
		      offset = prior;
		      goto jump;
		    }
		}
	      else if (c == 'B')
		{
		  bookmark = fopen ("RIP.MRK", "a");
		  if (bookmark != NULL)
		    {
		      fseek (bookmark, 0, SEEK_END);
		      fprintf (bookmark, "%lu %s\n", offset, infile);
		      fclose (bookmark);
		    }
		  goto done;
		}
	      else if ((num_search > 2 && (c == '+' || c == '-'))
		       || c == 'R' || c == 27 || c == 'Q' || c == 'X')
		goto done;
	    }
	}
    next:;
    }

done:
  textcolor (CLR_MENU);
  textbackground (BCK_MENU);
  lowvideo ();
  window (1, 1, 80, 25);
  gotoxy (1, 1);
  clrscr ();
  return (c);
}

/* Decompress and return the first n lines of the file */
void
summarize (FILE * fin, unsigned lines, int nonblank_only)
{
  int c;
  char *ss;
  init_oneline ();
  firstline = 0xffffffffL;
  current = 0;
  get_text (fin, num_text, 0, NULL, 0);
  for (; lines;)
    {
      c = get_oneline (fin);
      if (c == -1)
	break;
      if (nonblank_only)
	{
	  for (ss = s; isspace (*ss); ss++);
	  if (!*ss)
	    {
	      *s = 0;		/* continue; */
	    }
	}
      printf ("%s\n", s);
      if (*s)
	lines--;
    }
}


