/*
  Copyright 1996, 2002 Ronald S. Burkey.

  This file is part of Ron's Indexing Program (RIP).

  RIP is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  RIP is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GutenMark; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Filename:	ripc.c 
  Target:	MS-DOS
  Compiler:	Borland Turbo C 2.0
  Purpose:	RIP file-compression/file-indexing
  Mods:		10/11/96 RSB	Split off from RIP.C.
  		01/08/02 RSB	Reformatted in GNU style and otherwise
				cleaned up in preparation for first 
				GPL'd release.
*/

#include "rip.h"

/* While processing a text block, we maintain a binary tree of the hashcodes
   contained within the block.  At the end of the block, the values in the
   tree are output.  That way we have no duplicates. */
#define MAX_HASH_INBLOCK (OVERBLOCK/2)
typedef struct
{
  HASHVALUE hash;
  unsigned less, more;
}
HASH_INBLOCK;
HASH_INBLOCK hash_inblock[MAX_HASH_INBLOCK];
unsigned num_hash_inblock = 0;
void
add_hash_inblock (HASHVALUE hash)
{
  unsigned inblock_ptr;
  HASH_INBLOCK inblock;
  if (num_hash_inblock == 0)
    goto finish_add;
  else if (num_hash_inblock >= MAX_HASH_INBLOCK)
    return;
  else
    {
      inblock_ptr = 0;
      for (;;)
	{
	  inblock = hash_inblock[inblock_ptr];
	  if (hash < inblock.hash)
	    {
	      if (inblock.less)
		inblock_ptr = inblock.less;
	      else
		{
		  hash_inblock[inblock_ptr].less = num_hash_inblock;
		  goto finish_add;
		}
	    }
	  else if (hash > inblock.hash)
	    {
	      if (inblock.more)
		inblock_ptr = inblock.more;
	      else
		{
		  hash_inblock[inblock_ptr].more = num_hash_inblock;
		finish_add:
		  inblock.hash = hash;
		  inblock.less = 0;
		  inblock.more = 0;
		  hash_inblock[num_hash_inblock++] = inblock;
		  return;
		}
	    }
	  else
	    return;
	}
    }
}

extern FILE *ftemp, *ftemp2, *fin, *fout;
extern int delete_file;
extern COMPRESSOR *compressors[29];
extern DECOMPRESSOR *decompressors[29];

/* During the indexing process, completes the indexing for a block, and
   sets up for the next block.  The index must first be written out to
   a temporary file, though, so that it can be resorted.  */
int count;
long blockno;
void
reset_block (void)
{
  unsigned ub;
  int i;
#pragma warn -sig
  ub = blockno;
#pragma warn +sig
  for (i = 0; i < num_hash_inblock; i++)
    {
      fwrite (&ub, sizeof (ub), 1, ftemp);
      fwrite (&hash_inblock[i].hash, sizeof (HASHVALUE), 1, ftemp);
    }
  num_hash_inblock = 0;
  count = 0;
  blockno++;
  fseek (fin, blockno * BLOCKSIZE, SEEK_SET);
  do
    {
      i = getc (fin);
    }
  while (ISVALID (i) && i != EOF);
}

/* Reads a single character from a file, keeping a count */
int
rgetc (FILE * fp)
{
  int i;
  i = getc (fp);
  count++;
  return (toupper (i));
}

typedef struct
{
  unsigned block;
  HASHVALUE hash;
}
HASH_RECORD;

void longsort2 (int (*cmp) (void *, void *),
		FILE *, unsigned recsize, unsigned long, unsigned long);
int
cmp_hash (void *t1, void *t2)
{
#define h1 ((HASH_RECORD *) t1)
#define h2 ((HASH_RECORD *) t2)
  if (h1->hash < h2->hash)
    return (-1);
  else if (h1->hash > h2->hash)
    return (1);
  else if (h1->block < h2->block)
    return (-1);
  else if (h1->block > h2->block)
    return (1);
  else
    return (0);
#undef h1
#undef h2    
}

/* Index the open input file "fin" to the open output file "fout" */
int
rip (void)
{
  unsigned ub;
  HASHVALUE current_hash, uh, uhh;
  unsigned long n, nn, wordcount;
  int i, jj, x, y;
  char s[257];
  int SLEN;

  y = wherey ();
  x = wherex ();
  wordcount = 0;
  SLEN = sizeof (s) - 1;
  blockno = 0;
  num_hash_inblock = 0;
  fseek (ftemp, 0L, SEEK_SET);

  /* Pick off the hashcodes and write them out. */
  count = 0;
  blockno = 0;
  for (i = 0; i != EOF;)
    {
      jj = 0;
      s[0] = 0;
      i = rgetc (fin);
      if (count >= OVERBLOCK)
	reset_block ();
      if (isalpha (i))
	{
	  s[0] = i;
	  jj++;
	  for (;;)
	    {
	      i = rgetc (fin);
	      if (i == EOF)
		{
		endstring:
		  s[jj] = 0;
		  wordcount++;
		  if (!(wordcount & 0x3ff))
		    {
		      gotoxy (x, y);
		      cprintf (", %lu", wordcount);
		    }
		  if ((jj >= MINSTRING || !iswimp (s)))
		    add_hash_inblock (hash (s));
		  break;
		}
	      else
		{
		  if (ISVALID (i))
		    {
		      if (isalpha (i) && jj < SLEN)
			s[jj++] = i;
		    }
		  else
		    goto endstring;
		}
	    }
	}
    }
  if (count)
    reset_block ();

  /* Now the index for the file, which has been written out to a temporary
     file in the form of longs with the MSW equal to the hashcode and
     the LSW equal to the blocknumber, must be sorted. */
  gotoxy (x, y);
  cprintf (", Sorting");
  nn = ftell (ftemp) / sizeof (HASH_RECORD);
  fclose (ftemp);
  i = longsort4 (cmp_hash,
		 "RIP.TMP", sizeof (HASH_RECORD), 0, nn, "RIP3.TMP");
  if (i)
    return (i);
  ftemp = fopen ("RIP3.TMP", "r+b");
  setvbuf (ftemp, NULL, _IOFBF, 16384);

  /*
    ... and then copied back to the output file, in a somewhat altered form.
    The whole point of this exercise is that we'd like to read off just
    the block numbers for the specified hashcode without reading off
    all the others at the same time.  Here's the format:
     	word[]       	LS values of the sorted index table.  
			In other words, gives all the block numbers, 
			in increasing order, for the first hashcode, 
			then all the block numbers for the
       			second hashcode, etc.  These are not necessarily
       			hashcode=1, hashcode=2, etc., since not every hashcode
       			appears in the table.
     	HASHVALUE num_hashcodes;
     	struct {             	Table, in increasing order of hashcodes.
    	  HASHVALUE hashcode;   Gives pointers into the word[] data.
          unsigned long offset; Terminated by hashcode=0, offset=eof.
        } []
    In this form, we can jump right to the struct table, do a binary search
    on it to find where the specific hashcode data is, then jump right
    to that data.

    Anyhow, in order to do this, we write a lot of ftemp into fout, but
    put some in ftemp2.  The ftemp2 stuff we put into fout at the very
    end.
  */
  fseek (fout, 0, SEEK_END);
  n = ftell (fout);
  fseek (ftemp2, 0, SEEK_SET);
  fseek (ftemp, 0, SEEK_SET);
  for (current_hash = 0; nn; nn--)
    {
      if (1 != fread (&ub, sizeof (ub), 1, ftemp))
	break;
      if (1 != fread (&uh, sizeof (uh), 1, ftemp))
	break;
      if (uh != current_hash && current_hash != 0)
	{
	  fwrite (&current_hash, sizeof (current_hash), 1, ftemp2);
	  fwrite (&n, sizeof (n), 1, ftemp2);
	  n = ftell (fout);
	}
      current_hash = uh;
      fwrite (&ub, sizeof (ub), 1, fout);
    }
  fwrite (&current_hash, sizeof (current_hash), 1, ftemp2);
  fwrite (&n, sizeof (n), 1, ftemp2);
  current_hash = 0;
  n = ftell (fout);
  fwrite (&current_hash, sizeof (current_hash), 1, ftemp2);
  fwrite (&n, sizeof (n), 1, ftemp2);
#pragma warn -sig
  uhh = ftell (ftemp2) / (sizeof (unsigned long) + sizeof (HASHVALUE));
#pragma warn +sig
  fwrite (&uhh, sizeof (uhh), 1, fout);
  fseek (ftemp2, 0, SEEK_SET);
  for (; uhh; uhh--)
    {
      fread (&uh, sizeof (uh), 1, ftemp2);
      fread (&nn, sizeof (nn), 1, ftemp2);
      fwrite (&uh, sizeof (uh), 1, fout);
      fwrite (&nn, sizeof (nn), 1, fout);
    }
  fseek (fout, 12, SEEK_SET);
  fwrite (&n, sizeof (n), 1, fout);
  gotoxy (x, y);

  fclose (ftemp);
  remove ("RIP3.TMP");
  ftemp = fopen ("RIP.TMP", "r+b");
  setvbuf (ftemp, NULL, _IOFBF, 16384);
  return (0);
}

/* Create a rip-file */
int
compress_it (char *infile, char *outfile)
{
  char *sss;
  int err;
  unsigned long m, n, blocksize, index, bit_dummy, blocks, kb, block, o_size;
  unsigned i, k;
  struct ftime ft;
  COMPRESSOR *compress;
  int lastchar;
  unsigned long checksum;
  unsigned long t;

  t = clock ();
  lastchar = ' ';
  for (sss = infile; *sss && *sss != '.'; sss++);
  if (!strncmpi (sss, ".RIP", 4))
    return (3);

  for (sss = outfile; *sss && *sss != '.'; sss++);
  strcpy (sss, ".rip");

  err = open_file (infile);
  if (err)
    return (0x10 + err);
  err = create_file (outfile);
  if (err)
    return (0x20 + err);
  cprintf ("%s->RIP", infile);

  fprintf (fout, "RIP3");
  fseek (fin, 0, SEEK_END);
  o_size = n = ftell (fin);
  fseek (fin, 0, SEEK_SET);
  fwrite (&n, sizeof (n), 1, fout);
  blocksize = BLOCKSIZE;
  fwrite (&blocksize, sizeof (blocksize), 1, fout);
  blocks = (n + blocksize - 1) / blocksize;
  index = 0;
  fwrite (&index, sizeof (index), 1, fout);
  bit_dummy = (16 + 4 * blocks) * 8;
  for (n = 0; n < blocks; n++)
    {
      fwrite (&bit_dummy, sizeof (bit_dummy), 1, fout);
      bit_dummy = 0;
    }

  block = 0;			/* current block */
  kb = 0;			/* number of bytes written so far in this block */
  m = 0;			/* packed output bits */
  k = 0;			/* bit count. */
  for (;;)
    {
      i = getc (fin);
      if (((int) i) == EOF)
	break;
      if (isalpha (lastchar))
	{
	  compress = compressors[toupper (lastchar) - 'A'];
	}
      else if (isdigit (lastchar))
	{
	  compress = compressors[26];
	}
      else if (ispunctuation (lastchar))
	{
	  compress = compressors[27];
	}
      else
	{
	  compress = compressors[28];
	}
      lastchar = i;
      n = (*compress)[i].bits;
      i = (*compress)[i].numbits;
      n |= (m << i);
      i += k;
      while (i >= 8)
	{
	  i -= 8;
	  m = n >> i;
	  fwrite (&m, 1, 1, fout);
	}
      m = n;
      k = i;
      kb++;
      if (kb == BLOCKSIZE)
	{
	  lastchar = ' ';
	  kb = 0;
	  block++;
	  n = ftell (fout);
	  fseek (fout, 16 + 4 * block, SEEK_SET);
	  bit_dummy = 8 * n + k;
	  fwrite (&bit_dummy, sizeof (bit_dummy), 1, fout);
	  fseek (fout, n, SEEK_SET);
	}
    }
  if (k)
    {
      m = m << (8 - k);
      fwrite (&m, 1, 1, fout);
    }

  /* index it */
  index = ftell (fout);
  fseek (fout, 12, SEEK_SET);
  fwrite (&index, sizeof (index), 1, fout);
  fseek (fout, index, SEEK_SET);
  fseek (fin, 0, SEEK_SET);
  i = rip ();
  if (i)
    return (i);

  /* Add in a file checksum */
  fseek (fout, 8, SEEK_SET);
  checksum = 0;
  for (;;)
    {
      err = getc (fout);
      if (err == EOF)
	break;
      checksum += (unsigned) err;
    }
  fseek (fout, 0, SEEK_END);
  fwrite (&checksum, sizeof (checksum), 1, fout);

  fseek (fout, 0, SEEK_END);
  m = ftell (fout);
  cprintf (", %lu->%lu", o_size, m);
  getftime (fileno (fin), &ft);
  fflush (fout);
  setftime (fileno (fout), &ft);
  if (ferror (fout))
    return (5);
  else
    {
      fclose (fout);
      fclose (fin);
      if ( /* o_size>m && */ delete_file)
	remove (infile);
    }
  cprintf (", %lu ticks\r\n", clock () - t);
  return (0);
}

